

#include <windows.h>
#include "cv.h"
#include "highgui.h"
#include <stdio.h>		// printf() / scanf()
#define _USE_MATH_DEFINES
#include <math.h>
#include "filter.h"

#include <io.h>
#include <string>
#include <vector>
#include <map>

#ifndef _DEBUG
#pragma comment(lib,"cv.lib")
#pragma comment(lib,"cvaux.lib")
#pragma comment(lib,"cxcore.lib")
#pragma comment(lib,"highgui.lib")
#else
#pragma comment(lib,"cvd.lib")
#pragma comment(lib,"cvauxd.lib")
#pragma comment(lib,"cxcored.lib")
#pragma comment(lib,"highguid.lib")
#endif

#define YC2R(Y,Cb,Cr) (Y + 1.402*Cr)
#define YC2G(Y,Cb,Cr) (Y - 0.344*Cb - 0.714*Cr)
#define YC2B(Y,Cb,Cr) (Y + 1.772*Cb)
#define RGB2Y(R,G,B) (0.299*R + 0.587*G + 0.114*B)
#define RGB2CB(R,G,B) (-0.169*R - 0.331*G + 0.500*B)
#define RGB2CR(R,G,B) (0.500*R - 0.419*G - 0.081*B)

// [min, max] ͈̔͂ɖOa
template <class T>
T Saturate(T v, T min, T max) {
	if (v < min) {
		return min;
	} else if (v > max) {
		return max;
	}
	return v;
}

template <>
int Saturate(int v, int min, int max) {
    v -= min;
    v &= ~(v >> 31);
    v += min;

    v -= max;
    v &=  (v >> 31);
    v += max;
    
    return v;
}


void CopyBufToCvBGRImg(
		PIXEL_YC *buf, IplImage *img,
		int x0, int y0, int w, int h, int x1, int y1, int mw, int mh
){
	int src,ix0,iy0,ix1,iy1;

	for(iy0 = y0; iy0 < y0 + h; iy0++) {
		iy1 = y1 + (iy0-y0);
		src = x0 + iy0 * mw;
		for(ix0 = x0,ix1 = x1; ix0 < x0 + w; ix0++) {
			int y  = buf[src].y;
			int cb = buf[src].cb;
			int cr = buf[src].cr;

			int r = ( 255*y + ( ((22881*cr>>16)+3) <<10) ) >>12;
			int g = ( 255*y + ( ((-5616*cb>>16)+(-11655*cr>>16)+3) <<10) ) >>12;
			int b = ( 255*y + ( ((28919*cb>>16)+3) <<10) ) >>12;

			CV_IMAGE_ELEM(img, unsigned char, iy1, ix1*3+0) = Saturate(b, 0, 255);
			CV_IMAGE_ELEM(img, unsigned char, iy1, ix1*3+1) = Saturate(g, 0, 255);
			CV_IMAGE_ELEM(img, unsigned char, iy1, ix1*3+2) = Saturate(r, 0, 255);
			src++;
			ix1++;
		}
	}
}

void CopyCvBGRImgToBuf(
		IplImage *img, PIXEL_YC *buf, 
		int x0, int y0, int w, int h, int x1, int y1, int mw, int mh
){
	int dst,ix0,iy0,ix1,iy1;
	PIXEL_YC yc;

	for(iy0 = y0; iy0 < y0 + h; iy0++) {
		iy1 = y1 + (iy0-y0);
		dst = x1 + iy1 * mw;
		for(ix0 = x0,ix1 = x1; ix0 < x0 + w; ix0++) {
			int r,g,b;
			b = CV_IMAGE_ELEM(img, unsigned char, iy0, ix0*3+0);
			g = CV_IMAGE_ELEM(img, unsigned char, iy0, ix0*3+1);
			r = CV_IMAGE_ELEM(img, unsigned char, iy0, ix0*3+2);
			yc.y =  ((4918*r+354)>>10) + ((9655*g+585)>>10) + ((1875*b+523)>>10);
			yc.cb = ((-2775*r+240)>>10) + ((-5449*g+515)>>10) + (( 8224*b+256)>>10);
			yc.cr = ((8224*r+256)>>10) + ((-6887*g+110)>>10) + ((-1337*b+646)>>10);

			buf[dst] = yc;
			dst++;
			ix1++;
		}
	}
}


static inline int is_neg(int v) {
    return (((unsigned int)v) >> (sizeof(v)*8 - 1));
}

static inline int modulo(int a, int b) {
	int m = a % b;
//	return m + b * is_neg(m);

	if (m >= 0) {
		return m;
	} else {
		return m + b;
	}
}


static IplImage *img0 = NULL;
static IplImage *img1 = NULL;

static char *g_dllpath = NULL;
static std::map<std::string, int> g_trackMap;
static std::map<std::string, int> g_checkMap;
static std::vector<int> g_bef_check;

static CvSubdiv2D *g_subdiv;
static CvMemStorage *g_subdivStorage;

static int g_nDiv = 1000;
static int g_lineL = 0;
static CvRNG g_rngState;

static bool g_drawLine = false;

static double g_equality = 1.0;
static bool g_divEqual = false;


static int getTrackVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_trackMap.find(str);

	if (itr == g_trackMap.end()) {
		sprintf_s(errmsg, "track %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	return fp->track[itr->second];
}

static int getCheckVal(FILTER *fp, char *str) {
	char errmsg[8192];

	std::map<std::string, int>::iterator itr;
	itr = g_checkMap.find(str);

	if (itr == g_checkMap.end()) {
		sprintf_s(errmsg, "check %s ݂͑܂", str);
		MessageBox(fp->hwnd, errmsg, "G[", MB_OK);
		return 0;
	}

	return fp->check[itr->second];
}


static void createImages(CvSize size) {
	img0		= cvCreateImage(size, IPL_DEPTH_8U, 3);
	img1		= cvCreateImage(size, IPL_DEPTH_8U, 3);
}

static void releaseImages() {
	cvReleaseImage(&img0);
	cvReleaseImage(&img1);
}


void initVoronoi(char *dllpath, FILTER_DLL *fdp) {
	g_dllpath = dllpath;

	g_trackMap.clear();
	for (int i=0; i<fdp->track_n; i++) {
		g_trackMap[fdp->track_name[i]] = i;
	}

	g_checkMap.clear();
	for (int i=0; i<fdp->check_n; i++) {
		g_checkMap[fdp->check_name[i]] = i;
	}
	g_bef_check.resize(fdp->check_n, 0);
}

void updateVoronoi(FILTER *fp) {
}


void exitVoronoi() {
	releaseImages();
}


static CvSubdiv2D* init_delaunay( CvMemStorage* storage,
                           CvRect rect )
{
    CvSubdiv2D* subdiv;

    subdiv = cvCreateSubdiv2D( CV_SEQ_KIND_SUBDIV2D, sizeof(*subdiv),
                               sizeof(CvSubdiv2DPoint),
                               sizeof(CvQuadEdge2D),
                               storage );
    cvInitSubdivDelaunay2D( subdiv, rect );

    return subdiv;
}


static void draw_subdiv_point( IplImage* img, CvPoint2D32f fp, CvScalar color )
{
    cvCircle( img, cvPoint(cvRound(fp.x), cvRound(fp.y)), 3, color, CV_FILLED, 8, 0 );
}


static void draw_subdiv_edge( IplImage* img, CvSubdiv2DEdge edge, CvScalar color )
{
    CvSubdiv2DPoint* org_pt;
    CvSubdiv2DPoint* dst_pt;
    CvPoint2D32f org;
    CvPoint2D32f dst;
    CvPoint iorg, idst;

    org_pt = cvSubdiv2DEdgeOrg(edge);
    dst_pt = cvSubdiv2DEdgeDst(edge);

    if( org_pt && dst_pt )
    {
        org = org_pt->pt;
        dst = dst_pt->pt;

        iorg = cvPoint( cvRound( org.x ), cvRound( org.y ));
        idst = cvPoint( cvRound( dst.x ), cvRound( dst.y ));

        cvLine( img, iorg, idst, color, 1, CV_AA, 0 );
    }
}


static void draw_subdiv( IplImage* img, CvSubdiv2D* subdiv,
                  CvScalar delaunay_color, CvScalar voronoi_color )
{
    CvSeqReader  reader;
    int i, total = subdiv->edges->total;
    int elem_size = subdiv->edges->elem_size;

    cvStartReadSeq( (CvSeq*)(subdiv->edges), &reader, 0 );

    for( i = 0; i < total; i++ )
    {
        CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);

        if( CV_IS_SET_ELEM( edge ))
        {
            draw_subdiv_edge( img, (CvSubdiv2DEdge)edge + 1, voronoi_color );
            draw_subdiv_edge( img, (CvSubdiv2DEdge)edge, delaunay_color );
        }

        CV_NEXT_SEQ_ELEM( elem_size, reader );
    }
}

static void draw_subdiv_facet(IplImage *src, IplImage *dst, CvSubdiv2DEdge edge, int e)
{
    CvSubdiv2DEdge t = edge;
    int i, count = 0;
    CvPoint* buf = 0;
	CvPoint2D32f cp = {0.0, 0.0};

    // count number of edges in facet
    do
    {
        count++;
        t = cvSubdiv2DGetEdge( t, CV_NEXT_AROUND_LEFT );
    } while (t != edge );

    buf = (CvPoint*)malloc( count * sizeof(buf[0]));

    // gather points
    t = edge;
    for( i = 0; i < count; i++ )
    {
        CvSubdiv2DPoint* pt = cvSubdiv2DEdgeOrg( t );
        if( !pt ) break;

		cp.x += pt->pt.x;
		cp.y += pt->pt.y;

		buf[i] = cvPoint( cvRound(pt->pt.x), cvRound(pt->pt.y));
//draw_subdiv_point( dst, cvPoint2D32f(buf[i].x, buf[i].y), CV_RGB(0,255,255));/**/
        t = cvSubdiv2DGetEdge( t, CV_NEXT_AROUND_LEFT );
    }

    if( i == count )
    {
		cp.x /= count;
		cp.y /= count;
		int y = Saturate(cvRound(cp.y), 0, src->height - 1);
		int x = Saturate(cvRound(cp.x), 0, src->width  - 1);

		CvScalar col = CV_RGB(CV_IMAGE_ELEM(src, BYTE, y, x*3+2), 
							  CV_IMAGE_ELEM(src, BYTE, y, x*3+1),
							  CV_IMAGE_ELEM(src, BYTE, y, x*3+0));

//		CvScalar col = cvGet2D(src, Saturate(cvRound(cp.y), 0, src->height - 1), 
//									Saturate(cvRound(cp.x), 0, src->width  - 1));
        cvFillConvexPoly( dst, buf, count, col, CV_AA, 0 );
		if (g_drawLine) {
			cvPolyLine( dst, &buf, &count, 1, 1, CV_RGB(g_lineL,g_lineL,g_lineL), 1, CV_AA, 0);
		}
		/*
		if (e == 3) {
			CvPoint2D32f p = pt->pt;
			p.x += 5;
			draw_subdiv_point( dst, p, CV_RGB(0,255,255));
			p.x -= 5;
		} else {
			draw_subdiv_point( dst, pt->pt, CV_RGB(0,0,0));
		}
		*/
    }
    free( buf );
}

static void paint_voronoi(IplImage *src, IplImage *dst, CvSubdiv2D* subdiv)
{
    CvSeqReader  reader;
    int i, total = subdiv->edges->total;
    int elem_size = subdiv->edges->elem_size;

    cvStartReadSeq( (CvSeq*)(subdiv->edges), &reader, 0 );

    for( i = 0; i < total; i++ )
    {
        CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);

        if( CV_IS_SET_ELEM( edge ))
        {
            CvSubdiv2DEdge e = (CvSubdiv2DEdge)edge;
            // left
            draw_subdiv_facet( src, dst, cvSubdiv2DRotateEdge( e, 1 ), 1);

            // right
            draw_subdiv_facet( src, dst, cvSubdiv2DRotateEdge( e, 3 ), 3);
        }

        CV_NEXT_SEQ_ELEM( elem_size, reader );
    }
}


/*
static void drawVoronoi(IplImage *src, IplImage *dst)
{
    char win[] = "source";
    int i;
    CvRect rect = { 0, 0, src->width, src->height };
    CvMemStorage* storage;
    CvSubdiv2D* subdiv;
	CvScalar active_facet_color, delaunay_color, voronoi_color, bkgnd_color;

    active_facet_color = CV_RGB( 255, 0, 0 );
    delaunay_color  = CV_RGB( 0,0,0);
    voronoi_color = CV_RGB(0, 180, 0);
    bkgnd_color = CV_RGB(0,255,0);

    cvSet( dst, bkgnd_color, 0 );

    cvNamedWindow( win, 1 );

    storage = cvCreateMemStorage(0);
    subdiv = init_delaunay( storage, rect );

	CvPoint2D32f *fps = (CvPoint2D32f*)malloc(g_nDiv * sizeof(CvPoint2D32f)); 

    for( i = 0; i < g_nDiv; i++ ) {
		fps[i].x = (float)(cvRandReal(&g_rngState) * src->width);
		fps[i].y = (float)(cvRandReal(&g_rngState) * src->height);
        cvSubdivDelaunay2DInsert( subdiv, fps[i] );
    }
    
	cvCalcSubdivVoronoi2D( subdiv );
    cvSet( dst, bkgnd_color, 0 );
    paint_voronoi(src, dst, subdiv);
	
	free(fps);
	
	cvShowImage( win, dst );

    cvReleaseMemStorage( &storage );
}
*/


void releaseSubdiv() {
    cvReleaseMemStorage(&g_subdivStorage);
	g_subdiv = NULL;
}

void updateSubdiv(int width, int height) {
	releaseSubdiv();
	g_subdivStorage = cvCreateMemStorage(0);

	CvRect rect = {0, 0, width, height};
    g_subdiv = init_delaunay(g_subdivStorage, rect );

	if (g_divEqual) {
		double mgn = sqrt((double)(width * height)/g_nDiv);
		int xdiv = Saturate(cvRound(width/mgn), 0, width);
		int ydiv = Saturate(cvRound(height/mgn), 0, height);

		float ixdiv = 1.0f/(float)xdiv;
		float iydiv = 1.0f/(float)ydiv;

		float fw = (float)width;
		float fh = (float)height;
		float wd = fw*ixdiv;
		float hd = fh*iydiv;

		for (int j = 0; j <= ydiv; j++) {
			for (int i = 0; i <= xdiv; i++) {
				float d = (j % 2) * wd * 0.5f; 
				CvPoint2D32f pt = {wd*i + d, hd*j};
				pt.x += (float)(wd * (cvRandReal(&g_rngState) - 0.5) * (1.0-g_equality));
				pt.y += (float)(hd * (cvRandReal(&g_rngState) - 0.5) * (1.0-g_equality));
				pt.x = Saturate(pt.x, 0.0f, (float)(width-1));
				pt.y = Saturate(pt.y, 0.0f, (float)(height-1));
				cvSubdivDelaunay2DInsert(g_subdiv, pt);
			}
		}
	} else {
		for(int i = 0; i < g_nDiv; i++ ) {
			CvPoint2D32f pt;
			pt.x = (float)(cvRandReal(&g_rngState) * width);
			pt.y = (float)(cvRandReal(&g_rngState) * height);
			pt.x = Saturate(pt.x, 0.0f, (float)(width-1));
			pt.y = Saturate(pt.y, 0.0f, (float)(height-1));
			cvSubdivDelaunay2DInsert(g_subdiv, pt);
	    }
	}
    
	cvCalcSubdivVoronoi2D(g_subdiv);	
}

void drawSubdiv(IplImage *src, IplImage *dst) {
    CvScalar bkgnd_color = CV_RGB(0,255,0);
	cvSet( dst, bkgnd_color, 0 );
    paint_voronoi(src, dst, g_subdiv);
}




BOOL execVoronoi(FILTER *fp,FILTER_PROC_INFO *fpip) {
	PIXEL_YC *buffer0 = fpip->ycp_edit;
	PIXEL_YC *buffer1 = fpip->ycp_temp;

	int mw = fpip->max_w;
	int mh = fpip->max_h;
	int w = fpip->w;
	int h = fpip->h;

	bool updSubdiv = false;
	int	 b_nDiv = g_nDiv;

	g_nDiv = getTrackVal(fp, "");
	if (g_nDiv != b_nDiv) {
		updSubdiv = true;
	}

	g_lineL = getTrackVal(fp, "x");

	int freq = getTrackVal(fp, "XVԊu");
	int rseed = getTrackVal(fp, "");

	g_drawLine = (getCheckVal(fp, "E") == 1);


	double b_equality = g_equality;
	g_equality = (double)getTrackVal(fp, "ϓx")/100.0;
	if (g_equality != b_equality) {
		updSubdiv = true;
	}


	bool b_divEqual = g_divEqual;
	g_divEqual = (getCheckVal(fp, "ϓ") == 1);
	if (g_divEqual != b_divEqual) {
		updSubdiv = true;
	}


	// KvimagemۂȂ
	if (img0 == NULL) {
		releaseImages();
		createImages(cvSize(w,h));
		updSubdiv = true;
	} else {
		CvSize size = cvGetSize(img0);
		if (size.width != w || size.height != h) {
			releaseImages();
			createImages(cvSize(w,h));
			updSubdiv = true;
		}
	}

	// RNGStateXV
	static int64 seed = -1;
	int64 b_seed = seed;

	if (freq == 0) {
		seed = ((int64)(rseed+g_nDiv)) << 32;
	} else {
		seed = (((int64)(rseed+g_nDiv)) << 32) + fpip->frame / freq;
	}
	g_rngState = cvRNG(seed);

	if (seed != b_seed) {
		updSubdiv = true;
	}

	// {mC
	if (updSubdiv) {
		updateSubdiv(w, h);
	}


	//AviUtl  OpenCV
	CopyBufToCvBGRImg(buffer0, img0, 0,0,w,h, 0,0, mw,mh);

	//{mCʕ`
	drawSubdiv(img0, img1);

	//OpenCV  AviUtl
	CopyCvBGRImgToBuf(img1, buffer0, 0,0,w,h, 0,0, mw,mh);

	return TRUE;
}
